(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: VertexShader.cpp
//
// Desc: Example code showing how to do vertex shaders in D3D.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit VertexShaderUnit;

interface

uses
  Windows, Messages, Resource,
  Direct3D9, D3DX9, D3DApp, D3DEnumeration, D3DSettings,
  D3DFile, D3DFont, D3DUtil, DXUtil;

// C++Builder compile hack
{$NOINCLUDE Resource}
{$HPPEMIT '#include "Resource.hpp"'}

type
  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    // Font for drawing text
    m_pFont:            CD3DFont;
    m_pFontSmall:       CD3DFont;

    // Scene
    m_pVB:              IDirect3DVertexBuffer9;
    m_pIB:              IDirect3DIndexBuffer9;

    m_dwNumVertices:    DWord;
    m_dwNumIndices:     DWord;
    m_pVertexShader:    IDirect3DVertexShader9;
    m_pVertexDeclaration: IDirect3DVertexDeclaration9;
    m_dwSize:           DWord;

    // Transforms
    m_matPosition:      TD3DXMatrix;
    m_matView:          TD3DXMatrix;
    m_matProj:          TD3DXMatrix;

    // Navigation
    m_bKey:             array [0..255] of Byte;
    m_fSpeed:           Single;
    m_fAngularSpeed:    Single;
    m_bShowHelp:        Bool;

    m_vVelocity:        TD3DXVector3;
    m_vAngularVelocity: TD3DXVector3;

  protected
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function FinalCleanup: HResult; override;
    function Render: HResult; override;
    function FrameMove: HResult; override;
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
  public
    constructor Create;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM ): LRESULT; override;
  end;


implementation

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited;
  m_strWindowTitle   := 'VertexShader';
  m_d3dEnumeration.AppUsesDepthBuffer := True;

  m_pFont            := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pFontSmall       := CD3DFont.Create('Arial',  9, D3DFONT_BOLD);
  m_pIB              := nil;
  m_pVB              := nil;
  m_dwSize           := 32;
  m_dwNumIndices     := (m_dwSize - 1) * (m_dwSize - 1) * 6;
  m_dwNumVertices    := m_dwSize * m_dwSize;
  m_pVertexShader    := nil;
  m_pVertexDeclaration := nil;

  FillChar(m_bKey, SizeOf(m_bKey), 0);
  m_fSpeed           := 5.0;
  m_fAngularSpeed    := 1.0;
  m_bShowHelp        := False;

  m_vVelocity        := D3DXVector3Zero;
  m_vAngularVelocity := D3DXVector3Zero;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit;
var
  vEye, vAt, vUp: TD3DXVector3;
begin
  // Setup the view matrix
  vEye := D3DXVector3( 2.0, 3.0, 3.0 );
  vAt  := D3DXVector3( 0.0, 0.0, 0.0 );
  vUp  := D3DXVector3( 0.0, 1.0, 0.0 );
  D3DXMatrixLookAtRH( m_matView, vEye, vAt, vUp );

  // Set the position matrix
  D3DXMatrixInverse( m_matPosition, nil, m_matView );

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove;
var
  fSecsPerFrame: Single;
  vT, vR: TD3DXVector3;
  v1, v2: TD3DXVector3; // temporary

  matT, matR: TD3DXMatrix;
  qR: TD3DXQuaternion;

  mat: TD3DXMatrix;
  vA, vD: TD3DXVector4;
  vSin, vCos: TD3DXVector4;
begin
  fSecsPerFrame := m_fElapsedTime;

  // Process keyboard input
  vT:= D3DXVector3Zero;
  vR:= D3DXVector3Zero;

  if( m_bKey[VK_LEFT] or m_bKey[VK_NUMPAD1] ) <> 0       then vT.x:= vT.x - 1.0; // Slide Left
  if( m_bKey[VK_RIGHT] or m_bKey[VK_NUMPAD3] ) <> 0      then vT.x:= vT.x + 1.0; // Slide Right
  if( m_bKey[VK_DOWN] ) <> 0                             then vT.y:= vT.y - 1.0; // Slide Down
  if( m_bKey[VK_UP] ) <> 0                               then vT.y:= vT.y + 1.0; // Slide Up
  if( m_bKey[Ord('W')] ) <> 0                            then vT.z:= vT.z - 2.0; // Move Forward
  if( m_bKey[Ord('S')] ) <> 0                            then vT.z:= vT.z + 2.0; // Move Backward
  if( m_bKey[Ord('A')] or m_bKey[VK_NUMPAD8] ) <> 0      then vR.x:= vR.x - 1.0; // Pitch Down
  if( m_bKey[Ord('Z')] or m_bKey[VK_NUMPAD2] ) <> 0      then vR.x:= vR.x + 1.0; // Pitch Up
  if( m_bKey[Ord('E')] or m_bKey[VK_NUMPAD6] ) <> 0      then vR.y:= vR.y - 1.0; // Turn Right
  if( m_bKey[Ord('Q')] or m_bKey[VK_NUMPAD4] ) <> 0      then vR.y:= vR.y + 1.0; // Turn Left
  if( m_bKey[VK_NUMPAD9] ) <> 0                          then vR.z:= vR.z - 2.0; // Roll CW
  if( m_bKey[VK_NUMPAD7] ) <> 0                          then vR.z:= vR.z + 2.0; // Roll CCW

  // m_vVelocity        := m_vVelocity * 0.9 + vT * 0.1;
  D3DXVec3Add(m_vVelocity,
    D3DXVec3Scale(v1, m_vVelocity, 0.9)^,
    D3DXVec3Scale(v2, vT, 0.1)^);
  // m_vAngularVelocity := m_vAngularVelocity * 0.9 + vR * 0.1;
  D3DXVec3Add(m_vAngularVelocity,
    D3DXVec3Scale(v1, m_vAngularVelocity, 0.9)^,
    D3DXVec3Scale(v2, vR, 0.1)^);

  // Update position and view matricies
  // vT := m_vVelocity * fSecsPerFrame * m_fSpeed;
  D3DXVec3Scale(vT, m_vVelocity, fSecsPerFrame * m_fSpeed);
  // vR := m_vAngularVelocity * fSecsPerFrame * m_fAngularSpeed;
  D3DXVec3Scale(vR, m_vAngularVelocity, fSecsPerFrame * m_fAngularSpeed);

  D3DXMatrixTranslation( matT, vT.x, vT.y, vT.z);
  D3DXMatrixMultiply( m_matPosition, matT, m_matPosition );

  D3DXQuaternionRotationYawPitchRoll( qR, vR.y, vR.x, vR.z );
  D3DXMatrixRotationQuaternion( matR, qR );

  D3DXMatrixMultiply( m_matPosition, matR, m_matPosition );
  D3DXMatrixInverse( m_matView, nil, m_matPosition );
  m_pd3dDevice.SetTransform(D3DTS_VIEW, m_matView);

  // Set up the vertex shader constants
  {}
    D3DXMatrixMultiply(mat, m_matView, m_matProj);
    D3DXMatrixTranspose(mat, mat);

    vA:= D3DXVector4(Sin(m_fTime)*15.0, 0.0, 0.5, 1.0);
    vD:= D3DXVector4(D3DX_PI, 1.0/(2.0*D3DX_PI), 2.0*D3DX_PI, 0.05);

    // Taylor series coefficients for sin and cos
    vSin:= D3DXVector4(1.0, -1.0/6.0, 1.0/120.0, -1.0/5040.0);
    vCos:= D3DXVector4(1.0, -1.0/2.0, 1.0/ 24.0, -1.0/ 720.0);

    m_pd3dDevice.SetVertexShaderConstantF( 0, @mat,  4);
    m_pd3dDevice.SetVertexShaderConstantF( 4, @vA,   1);
    m_pd3dDevice.SetVertexShaderConstantF( 7, @vD,   1);
    m_pd3dDevice.SetVertexShaderConstantF(10, @vSin, 1);
    m_pd3dDevice.SetVertexShaderConstantF(11, @vCos, 1);
  {}

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render;
begin
  // Clear the backbuffer
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
                       $000000ff, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    m_pd3dDevice.SetVertexDeclaration(m_pVertexDeclaration);
    m_pd3dDevice.SetVertexShader(m_pVertexShader);
    m_pd3dDevice.SetStreamSource(0, m_pVB, 0, SizeOf(TD3DXVector2));
    m_pd3dDevice.SetIndices(m_pIB);
    m_pd3dDevice.DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, m_dwNumVertices,
                                      0, m_dwNumIndices div 3);

    // Output statistics
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

    if m_bShowHelp then 
    begin
      m_pFontSmall.DrawText(2, 40, D3DCOLOR_ARGB(255,255,255,255),
                            'Keyboard controls:');
      m_pFontSmall.DrawText(20, 60, D3DCOLOR_ARGB(255,255,255,255),
                            'Move'#10'Turn'#10'Pitch'#10'Slide'#10 +
                            'Help'#10'Change device'#10'Exit');
      m_pFontSmall.DrawText(210, 60, D3DCOLOR_ARGB(255,255,255,255),
                            'W,S'#10'E,Q'#10'A,Z'#10'Arrow keys'#10 +
                            'F1'#10'F2'#10'Esc');
    end else
    begin
      m_pFontSmall.DrawText(2, 40, D3DCOLOR_ARGB(255,255,255,255),
                            'Press F1 for help');
    end;

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
var
  declVS: array[0..1] of TD3DVertexElement9 =
  (
    (Stream: 0; Offset: 0; _Type: D3DDECLTYPE_FLOAT2; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_POSITION; UsageIndex: 0),
    ({D3DDECL_END})
  );
function CMyD3DApplication.InitDeviceObjects;
var
  strVertexShaderPath: array[0..511] of Char;
  pCode: ID3DXBuffer;
  dwFlags: DWORD;
begin
  m_pFont.InitDeviceObjects(m_pd3dDevice);
  m_pFontSmall.InitDeviceObjects(m_pd3dDevice);

  // Create vertex shader
  {}
    declVS[1] := D3DDECL_END;
    Result := m_pd3dDevice.CreateVertexDeclaration(@declVS, m_pVertexDeclaration);
    if FAILED(Result) then Exit;

    // Find the vertex shader file
    Result := DXUtil_FindMediaFileCb(strVertexShaderPath, SizeOf(strVertexShaderPath), 'Ripple.vsh');
    if FAILED(Result) then Exit;

    dwFlags := 0;
{$IFDEF DEBUG}
    dwFlags := dwFlags or D3DXSHADER_DEBUG;
{$ENDIF}
    // Assemble the vertex shader from the file
    Result := D3DXAssembleShaderFromFile(strVertexShaderPath, nil, nil, dwFlags, @pCode, nil);
    if FAILED(Result) then Exit;

    // Create the vertex shader
    Result := m_pd3dDevice.CreateVertexShader(PDWORD(pCode.GetBufferPointer), m_pVertexShader);
    pCode := nil;
    if FAILED(Result) then Exit;
  {}

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects;
var
  pIndices: PWord;
  y, x: DWord;
  pVertices: ^TD3DXVector2;

  fAspectRatio: Single;
begin
  m_pFont.RestoreDeviceObjects;
  m_pFontSmall.RestoreDeviceObjects;

  // Setup render states
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

  // Create index buffer
  {}
    Result:= m_pd3dDevice.CreateIndexBuffer(m_dwNumIndices*SizeOf(WORD),
                               0, D3DFMT_INDEX16, D3DPOOL_DEFAULT, m_pIB, nil);
    if FAILED(Result) then Exit;

    Result:= m_pIB.Lock(0, m_dwNumIndices*SizeOf(WORD), Pointer(pIndices), 0);
    if FAILED(Result) then Exit;

    for y:= 1 to m_dwSize - 1 do
    begin
      for x:= 1 to m_dwSize - 1 do
      begin
        pIndices^:= (y-1)*m_dwSize + (x-1); Inc(pIndices);
        pIndices^:= (y-0)*m_dwSize + (x-1); Inc(pIndices);
        pIndices^:= (y-1)*m_dwSize + (x-0); Inc(pIndices);

        pIndices^:= (y-1)*m_dwSize + (x-0); Inc(pIndices);
        pIndices^:= (y-0)*m_dwSize + (x-1); Inc(pIndices);
        pIndices^:= (y-0)*m_dwSize + (x-0); Inc(pIndices);
      end;
    end;

    Result:= m_pIB.Unlock;
    if FAILED(Result) then Exit;
  {}


  // Create vertex buffer
  {}
    Result:= m_pd3dDevice.CreateVertexBuffer(m_dwNumVertices*sizeof(TD3DXVector2),
                                             0, 0,
                                             D3DPOOL_DEFAULT, m_pVB, nil);
    if FAILED(Result) then Exit;

    Result:= m_pVB.Lock(0, m_dwNumVertices*SizeOf(TD3DXVector2), Pointer(pVertices), 0);
    if FAILED(Result) then Exit;

    for y:= 0 to m_dwSize - 1 do
    begin
      for x:= 0 to m_dwSize - 1 do
      begin
        // *pVertices++ = D3DXVECTOR2( ((float)x / (float)(m_dwSize-1) - 0.5f) * D3DX_PI,
        //                             ((float)y / (float)(m_dwSize-1) - 0.5f) * D3DX_PI );
        pVertices^ := D3DXVector2( (x / (m_dwSize-1) - 0.5) * D3DX_PI,
                                   (y / (m_dwSize-1) - 0.5) * D3DX_PI );
        Inc(pVertices);
      end;
    end;

    Result:= m_pVB.Unlock;
    if FAILED(Result) then Exit;
  {}

  // Set up the projection matrix
  fAspectRatio:= m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovRH(m_matProj, D3DXToRadian(60.0), fAspectRatio, 0.1, 100.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, m_matProj);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pFontSmall.InvalidateDeviceObjects;
  m_pIB:= nil;
  m_pVB:= nil;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects;
begin
  m_pFont.DeleteDeviceObjects;
  m_pFontSmall.DeleteDeviceObjects;
  SAFE_RELEASE(m_pVertexShader);
  SAFE_RELEASE(m_pVertexDeclaration);
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup;
begin
  m_pFont.Free;
  m_pFontSmall.Free;
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  if( (dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING) or
      (dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING)) <> 0 then 
  begin
    if (pCaps.VertexShaderVersion < D3DVS_VERSION(1,0)) then
    begin
      Result:= E_FAIL;
      Exit;
    end;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc;
begin
  // Handle key presses
  case uMsg of
    WM_KEYUP:   m_bKey[wParam]:= 0;
    WM_KEYDOWN: m_bKey[wParam]:= 1;
    WM_COMMAND:
    begin
      case LOWORD(wParam) of
        IDM_TOGGLEHELP: m_bShowHelp:= not m_bShowHelp;
      end;
    end;
  end;

  // Pass remaining messages to default handler
  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

end.

